from abc import ABC
from .identifiers import FQN


class ResourceScope(ABC):
    def fully_qualified_name(self, container, resource_name: str) -> FQN:
        raise NotImplementedError


class OrganizationScope(ResourceScope):
    def fully_qualified_name(self, _, resource_name: str) -> FQN:
        return FQN(name=resource_name)


class AccountScope(ResourceScope):
    def fully_qualified_name(self, _, resource_name: str) -> FQN:
        return FQN(name=resource_name)


class DatabaseScope(ResourceScope):
    def fully_qualified_name(self, database, resource_name: str) -> FQN:
        db = database.name if database else None
        return FQN(name=resource_name, database=db)


class SchemaScope(ResourceScope):
    def fully_qualified_name(self, schema, resource_name: str) -> FQN:
        db, sch = None, None
        if schema:
            db = schema.container.name if schema.container else None
            sch = schema.name if schema else None
        return FQN(name=resource_name, database=db, schema=sch)


class TableScope(ResourceScope):
    def fully_qualified_name(self, resource_name: str):
        raise NotImplementedError
        # return FQN(
        #     name=resource_name.upper(),
        #     database=self.database_name,
        #     schema=self.schema_name,
        #     table=self.table_name,
        # )



from dataclasses import dataclass

from .resource import Resource, ResourceContainer, ResourceSpec, ResourceNameTrait
from .role import Role
from .schema import Schema
from ..enums import ResourceType
from ..props import Props, IntProp, StringProp, TagsProp, FlagProp
from ..resource_name import ResourceName
from ..resource_tags import ResourceTags
from ..scope import AccountScope


@dataclass(unsafe_hash=True)
class _Database(ResourceSpec):
    name: ResourceName
    transient: bool = False
    owner: Role = "SYSADMIN"
    data_retention_time_in_days: int = 1
    max_data_extension_time_in_days: int = 14
    default_ddl_collation: str = None
    tags: ResourceTags = None
    comment: str = None


class Database(ResourceNameTrait, Resource, ResourceContainer):
    """
    Description:
        Represents a database in Snowflake.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-database

    Fields:
        name (string, required): The name of the database.
        transient (bool): Specifies if the database is transient. Defaults to False.
        owner (string or Role): The owner role of the database. Defaults to "SYSADMIN".
        data_retention_time_in_days (int): The number of days to retain data. Defaults to 1.
        max_data_extension_time_in_days (int): The maximum number of days to extend data retention. Defaults to 14.
        default_ddl_collation (string): The default collation for DDL statements.
        tags (dict): A dictionary of tags associated with the database.
        comment (string): A comment describing the database.

    Python:

        ```python
        database = Database(
            name="some_database",
            transient=True,
            owner="SYSADMIN",
            data_retention_time_in_days=7,
            max_data_extension_time_in_days=28,
            default_ddl_collation="utf8",
            tags={"project": "research", "priority": "high"},
            comment="This is a database."
        )
        ```

        A database can contain schemas. In Python, you can add a schema to a database in several ways:

        By database name:
        ```python
        sch = Schema(
            name = "some_schema",
            database = "my_test_db",
        )
        ```

        By database object:
        ```python
        db = Database(name = "my_test_db")
        sch = Schema(
            name = "some_schema",
            database = db,
        )
        ```

        Or using the `add` method:
        ```python
        db = Database(name = "my_test_db")
        sch = Schema(name = "some_schema")
        db.add(sch)
        ```

    Yaml:

        ```yaml
        databases:
          - name: some_database
            transient: true
            owner: SYSADMIN
            data_retention_time_in_days: 7
            max_data_extension_time_in_days: 28
            default_ddl_collation: utf8
            tags:
              project: research
              priority: high
            comment: This is a database.
        ```

        In yaml, you can add schemas to a database using the `schemas` field:

        ```yaml
        databases:
          - name: some_database
            schemas:
              - name: another_schema
        ```

        Or by name:

        ```yaml
        databases:
          - name: some_database
        schemas:
            - name: another_schema
              database: some_database
        ```
    """

    resource_type = ResourceType.DATABASE

    props = Props(
        transient=FlagProp("transient"),
        data_retention_time_in_days=IntProp("data_retention_time_in_days"),
        max_data_extension_time_in_days=IntProp("max_data_extension_time_in_days"),
        default_ddl_collation=StringProp("default_ddl_collation"),
        tags=TagsProp(),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _Database

    def __init__(
        self,
        name: str,
        transient: bool = False,
        owner: str = "SYSADMIN",
        data_retention_time_in_days: int = 1,
        max_data_extension_time_in_days: int = 14,
        default_ddl_collation: str = None,
        tags: dict[str, str] = None,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data = _Database(
            name=self._name,
            transient=transient,
            owner=owner,
            data_retention_time_in_days=data_retention_time_in_days,
            max_data_extension_time_in_days=max_data_extension_time_in_days,
            default_ddl_collation=default_ddl_collation,
            tags=tags,
            comment=comment,
        )
        if self._data.name != "SNOWFLAKE":
            self.add(
                Schema(name="PUBLIC", implicit=True),
                Schema(name="INFORMATION_SCHEMA", implicit=True),
            )
        self.requires(self._data.owner)

    def schemas(self):
        return self.items(resource_type=ResourceType.SCHEMA)

    @property
    def public_schema(self):
        return self.find(name="PUBLIC", resource_type=ResourceType.SCHEMA)



import json
import logging

from dataclasses import dataclass
from typing import Optional, Union
from queue import Queue

import snowflake.connector

from . import data_provider, lifecycle
from .builtins import SYSTEM_ROLES
from .client import (
    ALREADY_EXISTS_ERR,
    INVALID_GRANT_ERR,
    DOES_NOT_EXIST_ERR,
    execute,
    reset_cache,
)
from .diff import diff, Action
from .enums import ResourceType, ParseableEnum
from .identifiers import URN, FQN, resource_label_for_type
from .resources import Account, Database, Schema
from .resources.resource import Resource, ResourceContainer, ResourcePointer, convert_to_resource
from .resource_name import ResourceName
from .scope import AccountScope, DatabaseScope, OrganizationScope, SchemaScope

logger = logging.getLogger("titan")

SYNC_MODE_BLOCKLIST = [
    ResourceType.FUTURE_GRANT,
    ResourceType.GRANT,
    ResourceType.GRANT_ON_ALL,
    ResourceType.ROLE,
    ResourceType.USER,
    ResourceType.TABLE,
]


class MissingResourceException(Exception):
    pass


class MarkedForReplacementException(Exception):
    pass


class RunMode(ParseableEnum):
    CREATE_OR_UPDATE = "CREATE-OR-UPDATE"
    # FULLY_MANAGED = "FULLY-MANAGED"
    SYNC = "SYNC"
    SYNC_ALL = "SYNC-ALL"


@dataclass
class ResourceChange:
    action: Action
    urn: URN
    before: dict
    after: dict
    delta: dict

    def to_dict(self):
        return {
            "action": self.action.value,
            "urn": str(self.urn),
            "before": self.before,
            "after": self.after,
            "delta": self.delta,
        }


Manifest = dict[URN, dict]
State = dict[URN, dict]
Plan = list[ResourceChange]


def dump_plan(plan: Plan, format: str = "json"):
    if format == "json":
        return json.dumps([change.to_dict() for change in plan], indent=2)
    elif format == "text":
        """
        account:ABC123

        » role.transformer will be created

        + role "urn::ABC123:role/transformer" {
            + name  = "transformer"
            + owner = "SYSADMIN"
            }

        + warehouse "urn::ABC123:warehouse/transforming" {
            + name           = "transforming"
            + owner          = "SYSADMIN"
            + warehouse_type = "STANDARD"
            + warehouse_size = "LARGE"
            + auto_suspend   = 60
            }

        + grant "urn::ABC123:grant/..." {
            + priv = "USAGE"
            + on   = warehouse "transforming"
            + to   = role "transformer
            }

        + grant "urn::ABC123:grant/..." {
            + priv = "OPERATE"
            + on   = warehouse "transforming"
            + to   = role "transformer
            }
        """
        output = ""

        def _render_value(value):
            if isinstance(value, str):
                return f'"{value}"'
            return str(value)

        # green_start = "\033[36m"
        # color_end = "\033[0m"

        add_count = len([change for change in plan if change.action == Action.ADD])
        change_count = len([change for change in plan if change.action == Action.CHANGE])
        remove_count = len([change for change in plan if change.action == Action.REMOVE])

        output += "\n» titan[core]\n"
        output += f"» Plan: {add_count} to add, {change_count} to change, {remove_count} to destroy.\n\n"

        for change in plan:
            action_marker = ""
            if change.action == Action.ADD:
                action_marker = "+"
            elif change.action == Action.CHANGE:
                action_marker = "~"
            elif change.action == Action.REMOVE:
                action_marker = "-"
            output += f"{action_marker} {change.urn}"
            if change.action != Action.REMOVE:
                output += " {"
            output += "\n"
            key_lengths = [len(key) for key in change.delta.keys()]
            max_key_length = max(key_lengths) if len(key_lengths) > 0 else 0
            for key, value in change.delta.items():
                if key.startswith("_"):
                    continue
                new_value = _render_value(value)
                before_value = ""
                if key in change.before:
                    before_value = _render_value(change.before[key]) + " -> "
                output += f"  {action_marker} {key:<{max_key_length}} = {before_value}{new_value}\n"
            if change.action != Action.REMOVE:
                output += "}\n"
            output += "\n"
        return output
    else:
        raise Exception(f"Unsupported format {format}")


def print_plan(plan: Plan):
    print(dump_plan(plan, format="text"))


def plan_sql(plan: Plan) -> list[str]:
    sql_commands = []
    for change in plan:
        props = Resource.props_for_resource_type(change.urn.resource_type, change.after)
        if change.action == Action.ADD:
            sql_commands.append(lifecycle.create_resource(change.urn, change.after, props))
        elif change.action == Action.CHANGE:
            sql_commands.append(lifecycle.update_resource(change.urn, change.delta))
        elif change.action == Action.REMOVE:
            sql_commands.append(lifecycle.drop_resource(change.urn, change.before))
    return sql_commands


def print_diffs(diffs):
    for action, target, deltas in diffs:
        print(f"[{action}]", target)
        for delta in deltas:
            print("\t", delta)


def _split_by_scope(
    resources: list[Resource],
) -> tuple[list[Resource], list[Resource], list[Resource], list[Resource]]:
    org_scoped = []
    acct_scoped = []
    db_scoped = []
    schema_scoped = []

    def route(resource: Resource):
        """The sorting hat"""
        if isinstance(resource.scope, OrganizationScope):
            org_scoped.append(resource)
        elif isinstance(resource.scope, AccountScope):
            acct_scoped.append(resource)
        elif isinstance(resource.scope, DatabaseScope):
            db_scoped.append(resource)
        elif isinstance(resource.scope, SchemaScope):
            schema_scoped.append(resource)
        else:
            raise Exception(f"Unsupported resource type {type(resource)}")

    for resource in resources:
        route(resource)
        if isinstance(resource, ResourceContainer):
            for item in resource.items():
                route(item)
    return org_scoped, acct_scoped, db_scoped, schema_scoped


def _walk(resource: Resource):
    yield resource
    if isinstance(resource, ResourceContainer):
        for item in resource.items():
            yield from _walk(item)


def _raise_if_plan_would_drop_session_user(session_ctx: dict, plan: Plan):
    for change in plan:
        if change.urn.resource_type == ResourceType.USER and change.action == Action.REMOVE:
            if ResourceName(session_ctx["user"]) == ResourceName(change.urn.fqn.name):
                raise Exception("Plan would drop the current session user, which is not allowed")


class Blueprint:

    def __init__(
        self,
        name: str = None,
        resources: list[Resource] = None,
        run_mode: RunMode = RunMode.CREATE_OR_UPDATE,
        dry_run: bool = False,
        allowlist: list[ResourceType] = None,
        # account: Union[None, str, Account] = None,
        # database: Union[None, str, Database] = None,
        # schema: Union[None, str, Schema] = None,
        # allow_role_switching: bool = True,
        # ignore_ownership: bool = True,
        # valid_resource_types: List[ResourceType] = None,
        **kwargs,
    ) -> None:

        account = kwargs.pop("account", None)
        database = kwargs.pop("database", None)
        schema = kwargs.pop("schema", None)
        allow_role_switching = kwargs.pop("allow_role_switching", None)
        ignore_ownership = kwargs.pop("ignore_ownership", None)
        valid_resource_types = kwargs.pop("valid_resource_types", None)

        if account is not None:
            logger.warning("account is deprecated, use add instead")
        if database is not None:
            logger.warning("database is deprecated, use add instead")
        if schema is not None:
            logger.warning("schema is deprecated, use add instead")
        if allow_role_switching is not None:
            raise Exception("Role switching must be allowed")
        if ignore_ownership is not None:
            logger.warning("ignore_ownership is deprecated")
        if valid_resource_types is not None:
            logger.warning("valid_resource_types is deprecated")
            allowlist = valid_resource_types

        self._finalized = False
        self._staged: list[Resource] = []
        self._root: Account = None
        self._account_locator: str = None
        self._run_mode: RunMode = RunMode(run_mode)
        self._dry_run: bool = dry_run
        self._ignore_ownership: bool = ignore_ownership
        self._allowlist: list[ResourceType] = [ResourceType(v) for v in allowlist or []]

        if self._run_mode == RunMode.SYNC_ALL:
            raise Exception("Sync All mode is not supported yet")

        if ResourceType.USER in self._allowlist and self._run_mode != RunMode.CREATE_OR_UPDATE:
            raise Exception("User resource type is not allowed in this version of Titan")

        self.name = name
        self.account: Optional[Account] = convert_to_resource(Account, account) if account else None
        self.database: Optional[Database] = convert_to_resource(Database, database) if database else None
        self.schema: Optional[Schema] = convert_to_resource(Schema, schema) if schema else None

        if self.account and self.database:
            self.account.add(self.database)

        if self.database and self.schema:
            self.database.add(self.schema)

        self.add(resources or [])
        self.add([res for res in [self.account, self.database, self.schema] if res is not None])

    def _raise_for_nonconforming_plan(self, plan: Plan):
        exceptions = []

        # Run Mode exceptions
        if self._run_mode == RunMode.CREATE_OR_UPDATE:
            for change in plan:
                if change.action == Action.REMOVE:
                    exceptions.append(
                        f"Create-or-update mode does not allow resources to be removed (ref: {change.urn})"
                    )
                if change.action == Action.CHANGE:
                    if "owner" in change.delta:
                        change_debug = f"{change.before['owner']} => {change.delta['owner']}"
                        exceptions.append(
                            f"Create-or-update mode does not allow ownership changes (resource: {change.urn}, owner: {change_debug})"
                        )
                    elif "name" in change.delta:
                        exceptions.append(
                            f"Create-or-update mode does not allow renaming resources (ref: {change.urn})"
                        )

        if self._run_mode == RunMode.SYNC:
            for change in plan:
                if change.urn.resource_type in SYNC_MODE_BLOCKLIST:
                    exceptions.append(
                        f"Sync mode does not allow changes to {change.urn.resource_type} (ref: {change.urn})"
                    )

        # Valid Resource Types exceptions
        if self._allowlist:
            for change in plan:
                if change.urn.resource_type not in self._allowlist:
                    exceptions.append(f"Resource type {change.urn.resource_type} not allowed in blueprint")

        if exceptions:
            if len(exceptions) > 5:
                exception_block = "\n".join(exceptions[0:5]) + f"\n... and {len(exceptions) - 5} more"
            else:
                exception_block = "\n".join(exceptions)
            raise Exception("Non-conforming actions found in plan:\n" + exception_block)

    def _plan(self, remote_state: State, manifest: Manifest) -> Plan:
        manifest = manifest.copy()
        refs = manifest.pop("_refs")
        urns = manifest.pop("_urns")

        changes: Plan = []
        marked_for_replacement = set()
        for action, urn, delta in diff(remote_state, manifest):
            before = remote_state.get(urn, {})
            after = manifest.get(urn, {})

            if action == Action.CHANGE:
                if urn in marked_for_replacement:
                    continue

                resource_cls = Resource.resolve_resource_cls(urn.resource_type, before)

                # TODO: if the attr is marked as must_replace, then instead we yield a rename, add, remove
                attr = list(delta.keys())[0]
                attr_metadata = resource_cls.spec.get_metadata(attr)
                if attr_metadata.get("triggers_replacement", False):
                    raise MarkedForReplacementException(f"Resource {urn} is marked for replacement", resource_cls, attr)
                    marked_for_replacement.add(urn)
                elif attr_metadata.get("forces_add", False):
                    changes.append(ResourceChange(action=Action.ADD, urn=urn, before={}, after=after, delta=delta))
                    continue
                elif attr_metadata.get("fetchable", True) is False:
                    # drift on fields that aren't fetchable should be ignored
                    # TODO: throw a warning, or have a blueprint runmode that fails on this
                    continue
                elif attr == "owner" and self._ignore_ownership:
                    continue
                else:
                    changes.append(ResourceChange(action, urn, before, after, delta))
            elif action == Action.ADD:
                changes.append(ResourceChange(action=action, urn=urn, before={}, after=after, delta=delta))
            elif action == Action.REMOVE:
                changes.append(ResourceChange(action=action, urn=urn, before=before, after={}, delta={}))

        for urn in marked_for_replacement:
            raise MarkedForReplacementException(f"Resource {urn} is marked for replacement")
            # changes.append(ResourceChange(action=Action.REMOVE, urn=urn, before=before, after={}, delta={}))
            # changes.append(ResourceChange(action=Action.ADD, urn=urn, before={}, after=after, delta=after))

        # # Handle account role ownership by creating an ownership grant
        # for change in changes:
        #     if change.action == Action.ADD and change.after.get("owner") in SYSTEM_ROLES:
        #         changes.append(ResourceChange(action=Action.ADD, urn=change.urn, before={}, after=after, delta=after))

        # Generate a list of all URNs
        resource_set = set(urns + list(remote_state.keys()))
        for ref in refs:
            resource_set.add(ref[0])
            resource_set.add(ref[1])
        # Calculate a topological sort order for the URNs
        sort_order = topological_sort(resource_set, set(refs))
        return sorted(changes, key=lambda change: sort_order[change.urn])

    def fetch_remote_state(self, session, manifest: Manifest) -> State:
        state: State = {}

        manifest_urns: set[URN] = set(manifest["_urns"].copy())

        def _normalize(urn: URN, data: dict) -> dict:
            resource_cls = Resource.resolve_resource_cls(urn.resource_type, data)
            if urn.resource_type == ResourceType.FUTURE_GRANT:
                normalized = data
            elif isinstance(data, list):
                raise Exception(f"Fetching list of {urn.resource_type} is not supported yet")
                normalized = [resource_cls.defaults() | d for d in data]
            else:
                normalized = resource_cls.defaults() | data
            return normalized

        if self._run_mode == RunMode.SYNC:
            raise Exception("Sync mode is not supported yet")

        if self._run_mode == RunMode.SYNC_ALL:
            """
            In sync-all mode, the remote state is not just the resources that were added to the blueprint,
            but all resources that exist in Snowflake. This is limited by a few things:
            - allowlist limits the scope of what resources types are allowed in a blueprint
            - if database or schema is set, the blueprint only looks at that database or schema
            """
            if len(self._allowlist) == 0:
                raise Exception("Sync All mode must specify an allowlist")

            for resource_type in self._allowlist:
                for fqn in data_provider.list_resource(session, resource_label_for_type(resource_type)):
                    urn = URN(resource_type=resource_type, fqn=fqn, account_locator=self._account_locator)
                    data = data_provider.fetch_resource(session, urn)
                    if data is not None:
                        normalized_data = _normalize(urn, data)
                        state[urn] = normalized_data

        for urn in manifest.keys():
            if str(urn).startswith("_"):
                continue

            manifest_urns.remove(urn)
            data = data_provider.fetch_resource(session, urn)
            if data is not None:
                normalized_data = _normalize(urn, data)
                state[urn] = normalized_data

        # check for existence of resource refs
        for parent, reference in manifest["_refs"]:
            if reference in manifest:
                continue

            try:
                data = data_provider.fetch_resource(session, reference)
            except Exception as e:
                data = None
            if data is None:
                logger.error(manifest)
                raise MissingResourceException(
                    f"Resource {reference} required by {parent} not found or failed to fetch"
                )

        return state

    def _finalize(self, session_context: dict):
        """
        Convert the staged resources into a tree of resources
        """
        if self._finalized:
            return
        self._finalized = True

        org_scoped, acct_scoped, db_scoped, schema_scoped = _split_by_scope(self._staged)
        self._staged = None

        if len(org_scoped) > 1:
            raise Exception("Only one account allowed")
        elif len(org_scoped) == 1:
            # If we have a staged account, use it
            self._root = org_scoped[0]
        else:
            # Otherwise, create a stub account from the session context
            self._root = ResourcePointer(name=session_context["account"], resource_type=ResourceType.ACCOUNT)
            self._account_locator = session_context["account_locator"]
            self.account = self._root

        # Add all databases and other account scoped resources to the root
        for resource in acct_scoped:
            self._root.add(resource)

        # If we haven't specified a database, use the one from the session context
        if self.database is None and session_context.get("database") is not None:
            existing_databases = [db.name for db in self._root.items(resource_type=ResourceType.DATABASE)]
            if session_context["database"] not in existing_databases:
                self._root.add(ResourcePointer(name=session_context["database"], resource_type=ResourceType.DATABASE))

        databases: list[Database] = self._root.items(resource_type=ResourceType.DATABASE)

        # Add all schemas and database roles to their respective databases
        for resource in db_scoped:
            # When the container is present, we SHOULD be looking up the real container and attaching. In other words
            # this resource could be an island
            if resource.container is None:
                if len(databases) == 1:
                    databases[0].add(resource)
                else:
                    raise Exception(f"Database [{resource.container}] for resource {resource} not found")
            else:
                for db in databases:
                    if db.name == resource.container.name and resource not in db:
                        db.add(resource)
                        break

        for resource in schema_scoped:
            if resource.container is None:
                if len(databases) == 1:
                    public_schema: Schema = databases[0].find(resource_type=ResourceType.SCHEMA, name="PUBLIC")
                    if public_schema:
                        public_schema.add(resource)
                else:
                    raise Exception(f"No schema for resource {repr(resource)} found")
            elif isinstance(resource.container, ResourcePointer):
                schema_pointer = resource.container

                # We have a schema-scoped resource (eg a view) that has a resource pointer for the schema. The job is to connect
                # that resource into the tree
                #
                # If the schema pointer has no database, assume it lives in the only database we have
                if schema_pointer.container is None:
                    if len(databases) == 1:
                        databases[0].add(schema_pointer)
                    else:
                        raise Exception(f"No database for resource {resource} schema={resource.container}")
                elif isinstance(schema_pointer.container, ResourcePointer):
                    database_pointer = schema_pointer.container
                    found = False
                    for db in databases:
                        if db.name == database_pointer.name:
                            for schema in db.items(resource_type=ResourceType.SCHEMA):
                                if schema.name == schema_pointer.name:
                                    schema.add(resource)
                                    found = True
                                    break

                    if not found:
                        self._root.add(database_pointer)

            for ref in resource.refs:
                if ref.container is None and isinstance(ref.scope, resource.scope.__class__):
                    resource.container.add(ref)

        for database in databases:
            merge_schemas = []
            public_schema = None
            information_schema = None
            all_schemas = database.items(resource_type=ResourceType.SCHEMA)
            for schema in all_schemas:
                if schema.implicit:
                    if schema.name == "PUBLIC":
                        public_schema = schema
                    elif schema.name == "INFORMATION_SCHEMA":
                        information_schema = schema
                elif schema.name in ("PUBLIC", "INFORMATION_SCHEMA"):
                    merge_schemas.append(schema)

            for schema in merge_schemas:
                database.remove(schema)
                schema_items = schema.items()
                for item in schema_items:
                    schema.remove(item)
                    if schema.name == "PUBLIC":
                        public_schema.add(item)
                    elif schema.name == "INFORMATION_SCHEMA":
                        information_schema.add(item)

    def generate_manifest(self, session_context: dict = {}) -> Manifest:
        manifest: Manifest = {}
        refs = []
        urns = []

        self._finalize(session_context)

        for resource in _walk(self._root):
            if isinstance(resource, Resource):
                if resource.implicit and self._run_mode not in (RunMode.SYNC_ALL, RunMode.SYNC):
                    continue

            urn = URN(
                resource_type=resource.resource_type,
                fqn=resource.fqn,
                account_locator=self._account_locator,
            )

            data = resource.to_dict()

            if isinstance(resource, ResourcePointer):
                data["_pointer"] = True

            # manifest_key = str(urn)

            #### Special Cases
            if resource.resource_type == ResourceType.FUTURE_GRANT:
                # Role up FUTURE GRANTS on the same role/target to a single entry
                # TODO: support grant option, use a single character prefix on the priv
                if urn not in manifest:
                    manifest[urn] = {}
                on_type = data["on_type"].lower()
                if on_type not in manifest[urn]:
                    manifest[urn][on_type] = []
                if data["priv"] in manifest[urn][on_type]:
                    # raise Exception(f"Duplicate resource {urn} with conflicting data")
                    continue
                manifest[urn][on_type].append(data["priv"])

            #### Normal Case
            else:
                if urn in manifest:
                    if data != manifest[urn]:
                        # raise Exception(f"Duplicate resource {urn} with conflicting data")
                        continue
                manifest[urn] = data

            urns.append(urn)

            for ref in resource.refs:
                ref_urn = URN.from_resource(account_locator=self._account_locator, resource=ref)
                refs.append((urn, ref_urn))
        manifest["_refs"] = refs
        manifest["_urns"] = urns
        return manifest

    def plan(self, session) -> Plan:
        reset_cache()
        session_ctx = data_provider.fetch_session(session)
        manifest = self.generate_manifest(session_ctx)
        remote_state = self.fetch_remote_state(session, manifest)
        try:
            completed_plan = self._plan(remote_state, manifest)
        except Exception as e:
            logger.error("~" * 80 + "REMOTE STATE")
            logger.error(remote_state)
            logger.error("~" * 80 + "MANIFEST")
            logger.error(manifest)

            raise e
        self._raise_for_nonconforming_plan(completed_plan)
        return completed_plan

    def apply(self, session, plan: Plan = None):
        if plan is None:
            plan = self.plan(session)

        # TODO: cursor setup, including query tag
        # TODO: clean up urn vs urn_str madness

        """
            At this point, we have a list of actions as a part of the plan. Each action is one of:
                1. [ADD] action (CREATE command)
                2. [CHANGE] action (one or many ALTER or SET PARAMETER commands)
                3. [REMOVE] action (DROP command, REVOKE command, or a rename operation)

            Each action requires:
                • a set of privileges necessary to run commands
                • the appropriate role to execute commands

            Once we've determined those things, we can compare the list of required roles and privileges
            against what we have access to in the session and the role tree.
        """

        session_ctx = data_provider.fetch_session(session)

        _raise_if_plan_would_drop_session_user(session_ctx, plan)

        # required_privs = _collect_required_privs(session_ctx, plan)
        # available_privs = _collect_available_privs(session_ctx, session, plan, usable_roles)
        # _raise_if_missing_privs(required_privs, available_privs)

        action_queue = self._compile_plan_to_sql(session_ctx, plan)
        actions_taken = []

        while action_queue:
            sql = action_queue.pop(0)
            actions_taken.append(sql)
            try:
                if not self._dry_run:
                    execute(session, sql)
            except snowflake.connector.errors.ProgrammingError as err:
                if err.errno == ALREADY_EXISTS_ERR:
                    logger.error(f"Resource already exists: {sql}, skipping...")
                elif err.errno == INVALID_GRANT_ERR:
                    logger.error(f"Invalid grant: {sql}, skipping...")
                elif err.errno == DOES_NOT_EXIST_ERR and sql.startswith("REVOKE"):
                    logger.error(f"Resource does not exist: {sql}, skipping...")
                else:
                    raise err
        return actions_taken

    def _compile_plan_to_sql(self, session_ctx, plan: Plan):
        action_queue = []
        default_role = session_ctx["role"]

        def _queue_change(change: ResourceChange):
            """
            In Snowflake's RBAC model, a session has an active role, and zero or more secondary roles.

            The active role of a session is set as follows:
            - When a session is started:
                - If the session is configured with a role, that is the active role
                - Otherwise, if the user of the session has a default_role set, and that role exists, that is the active role
                - Otherwise, the PUBLIC role is activated (PUBLIC cannot be revoked)
            - Any time the USE ROLE command is run, the active role is switched

            A session may run any command thats allowed by the active role or any role downstream from it in the role hierarchy.
            When secondary roles are active (by running the command USE SECONDARY ROLES ALL), then the session may also run any
            command that any secondary role or a role downstream from it is allowed to run.

            However, when a CREATE command is run, only the active role is considered. This is because the role that
            creates a new resource owns that resource by default. There are some exceptions with GRANTS.

            For those reasons, we generally don't have to worry about the current role as long as we have activated secondary roles.
            The exception is when creating new resources
            """

            before_action = []
            action = None
            after_action = []
            if change.action == Action.ADD:
                props = Resource.props_for_resource_type(change.urn.resource_type, change.after)
                # TODO: raise exception if role isn't usable. Maybe can just let this fail naturally

                if "owner" in change.after:
                    if change.after["owner"] in SYSTEM_ROLES:
                        before_action.append(f"USE ROLE {change.after['owner']}")
                    else:
                        before_action.append(f"USE ROLE {default_role}")
                        after_action.append(
                            f"GRANT OWNERSHIP ON {change.urn.resource_type} {change.urn.fqn} TO {change.after['owner']}"
                        )
                elif change.urn.resource_type in (ResourceType.FUTURE_GRANT, ResourceType.ROLE_GRANT):
                    # TODO: switch to role with MANAGE GRANTS if we dont have access to SECURITYADMIN
                    before_action.append(f"USE ROLE SECURITYADMIN")
                else:
                    before_action.append(f"USE ROLE {default_role}")

                action = lifecycle.create_resource(change.urn, change.after, props)
            elif change.action == Action.CHANGE:
                action = lifecycle.update_resource(change.urn, change.delta)
            elif change.action == Action.REMOVE:
                action = lifecycle.drop_resource(change.urn, change.before, if_exists=True)

            action_queue.extend(before_action)
            action_queue.append(action)
            action_queue.extend(after_action)

        action_queue.append("USE SECONDARY ROLES ALL")
        for change in plan:
            _queue_change(change)
        return action_queue

    def destroy(self, session, manifest: Manifest = None):
        session_ctx = data_provider.fetch_session(session)
        manifest = manifest or self.generate_manifest(session_ctx)
        for urn, data in manifest.items():
            if str(urn).startswith("_"):
                continue

            if isinstance(data, dict) and data.get("_pointer"):
                continue
            if urn.resource_type == ResourceType.GRANT:
                for grant in data:
                    execute(session, lifecycle.drop_resource(urn, grant))
            else:
                try:
                    execute(session, lifecycle.drop_resource(urn, data))
                except snowflake.connector.errors.ProgrammingError as err:
                    continue

    def _add(self, resource: Resource):
        if self._finalized:
            raise Exception("Cannot add resources to a finalized blueprint")
        if not isinstance(resource, Resource):
            raise Exception(f"Expected a Resource, got {type(resource)} -> {resource}")
        self._staged.append(resource)

    def add(self, *resources):
        if isinstance(resources[0], list):
            resources = resources[0]
        for resource in resources:
            self._add(resource)


def topological_sort(resource_set: set, references: set):
    # Kahn's algorithm

    # Compute in-degree (# of inbound edges) for each node
    in_degrees = {}
    outgoing_edges = {}

    for node in resource_set:
        in_degrees[node] = 0
        outgoing_edges[node] = set()

    for node, ref in references:
        in_degrees[ref] += 1
        outgoing_edges[node].add(ref)

    # Put all nodes with 0 in-degree in a queue
    queue = Queue()
    for node, in_degree in in_degrees.items():
        if in_degree == 0:
            queue.put(node)

    # Create an empty node list
    nodes = []

    while not queue.empty():
        node = queue.get()
        nodes.append(node)

        # For each of node's outgoing edges
        empty_neighbors = set()
        for edge in outgoing_edges[node]:
            in_degrees[edge] -= 1
            if in_degrees[edge] == 0:
                queue.put(edge)
                empty_neighbors.add(edge)

        # Remove edges to empty neighbors
        outgoing_edges[node].difference_update(empty_neighbors)
    nodes.reverse()
    if len(nodes) != len(resource_set):
        raise Exception("Graph is not a DAG")
    return {value: index for index, value in enumerate(nodes)}



import yaml


from github import Github, InputGitTreeElement

from .identifiers import ResourceLocator, FQN
from .search import crawl_resources

CREATE_MODE = "100644"


def _git_path_for_resource(resource):
    if resource["resource_key"] == "database":
        return f"{resource['resource_key']}:{resource['name']}.yaml"
    elif resource["resource_key"] == "schema":
        return f"{resource['database']}/{resource['resource_key']}:{resource['name']}.yaml"
    else:
        return "UNKNOWN"


def export(session, repo: str, path: str, locator_str: str, access_token: str):
    """
    Export resources from a Snowflake account to a GitHub repository.

        repo (str): The name of the repository to import from. Ex: "teej/titan"
    """
    locator = ResourceLocator.from_str(locator_str)

    gh = Github(access_token)
    repo = gh.get_repo(repo)

    changes = []
    for resource in crawl_resources(session, locator):
        resource_path = _git_path_for_resource(resource)
        changes.append({"path": path + "/" + resource_path, "content": yaml.dump(resource)})

    # Create a tree element for each file change
    tree_elements = []
    for file_change in changes:
        blob = repo.create_git_blob(file_change["content"], "utf-8")
        tree_elements.append(InputGitTreeElement(path=file_change["path"], mode=CREATE_MODE, type="blob", sha=blob.sha))

    # Create a tree
    tree = repo.create_git_tree(tree_elements)

    # Get the current commit (head commit of the branch)
    parent = repo.get_git_ref("heads/main").object.sha
    parent_commit = repo.get_git_commit(parent)

    # Create a new commit
    new_commit = repo.create_git_commit("commit msg", tree, [parent_commit])

    # Update the reference to point to the new commit
    repo.get_git_ref("heads/main").edit(new_commit.sha)



from dataclasses import dataclass, field

from .resource import Resource, ResourceSpec, ResourceNameTrait
from .role import Role
from ..enums import ParseableEnum, ResourceType
from ..scope import SchemaScope
from ..resource_name import ResourceName
from ..props import (
    BoolProp,
    EnumProp,
    IdentifierProp,
    Props,
    PropSet,
    StringProp,
    TagsProp,
)


class StageType(ParseableEnum):
    INTERNAL = "INTERNAL"
    EXTERNAL = "EXTERNAL"


class EncryptionType(ParseableEnum):
    SNOWFLAKE_FULL = "SNOWFLAKE_FULL"
    SNOWFLAKE_SSE = "SNOWFLAKE_SSE"
    AWS_CSE = "AWS_CSE"
    AWS_SSE_S3 = "AWS_SSE_S3"
    AWS_SSE_KMS = "AWS_SSE_KMS"
    GCS_SSE_KMS = "GCS_SSE_KMS"
    AZURE_CSE = "AZURE_CSE"
    NONE = "NONE"


# """
# copyOptions ::=
#      ON_ERROR = { CONTINUE | SKIP_FILE | SKIP_FILE_<num> | 'SKIP_FILE_<num>%' | ABORT_STATEMENT }
#      SIZE_LIMIT = <num>
#      PURGE = TRUE | FALSE
#      RETURN_FAILED_ONLY = TRUE | FALSE
#      MATCH_BY_COLUMN_NAME = CASE_SENSITIVE | CASE_INSENSITIVE | NONE
#      ENFORCE_LENGTH = TRUE | FALSE
#      TRUNCATECOLUMNS = TRUE | FALSE
#      FORCE = TRUE | FALSE
# """
# copy_options = Props(
#     on_error=StringProp("on_error", alt_tokens=["CONTINUE", "SKIP_FILE", "ABORT_STATEMENT"]),
#     size_limit=IntProp("size_limit"),
#     purge=BoolProp("purge"),
#     return_failed_only=BoolProp("return_failed_only"),
#     match_by_column_name=StringProp("match_by_column_name", alt_tokens=["CASE_SENSITIVE", "CASE_INSENSITIVE", "NONE"]),
#     enforce_length=BoolProp("enforce_length"),
#     truncatecolumns=BoolProp("truncatecolumns"),
#     force=BoolProp("force"),
# )


@dataclass(unsafe_hash=True)
class _InternalStage(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    type: StageType = StageType.INTERNAL
    encryption: dict[str, EncryptionType] = field(default_factory=None, metadata={"fetchable": False})
    directory: dict[str, bool] = None
    tags: dict[str, str] = None
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.type != StageType.INTERNAL:
            raise ValueError("Type must be INTERNAL for _InternalStage")
        if self.encryption:
            if "type" not in self.encryption:
                raise ValueError("When specifying encryption, 'type' is required")
            if self.encryption["type"] not in [EncryptionType.SNOWFLAKE_FULL, EncryptionType.SNOWFLAKE_SSE]:
                raise ValueError("Encryption 'type' must be SNOWFLAKE_FULL or SNOWFLAKE_SSE for InternalStage")


class InternalStage(ResourceNameTrait, Resource):
    """
    -- Internal stage
    CREATE [ OR REPLACE ] [ { TEMP | TEMPORARY } ] STAGE [ IF NOT EXISTS ] <internal_stage_name>
        internalStageParams
        directoryTableParams
        [ COMMENT = '<string_literal>' ]
        [ [ WITH ] TAG ( <tag_name> = '<tag_value>' [ , <tag_name> = '<tag_value>' , ... ] ) ]

    internalStageParams ::=
      [ ENCRYPTION = (TYPE = 'SNOWFLAKE_FULL' | TYPE = 'SNOWFLAKE_SSE') ]

    directoryTableParams (for internal stages) ::=
      [ DIRECTORY = ( ENABLE = { TRUE | FALSE }
                      [ REFRESH_ON_CREATE =  { TRUE | FALSE } ] ) ]
    """

    resource_type = ResourceType.STAGE
    props = Props(
        encryption=PropSet(
            "encryption",
            Props(type=EnumProp("type", [EncryptionType.SNOWFLAKE_FULL, EncryptionType.SNOWFLAKE_SSE], quoted=True)),
        ),
        directory=PropSet(
            "directory", Props(enable=BoolProp("ENABLE"), refresh_on_create=BoolProp("REFRESH_ON_CREATE"))
        ),
        comment=StringProp("comment"),
        tags=TagsProp(),
    )
    scope = SchemaScope()
    spec = _InternalStage

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        encryption: dict[str, EncryptionType] = None,
        directory: dict[str, bool] = None,
        tags: dict[str, str] = None,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data: _InternalStage = _InternalStage(
            name=self._name,
            owner=owner,
            encryption=encryption,
            directory=directory,
            tags=tags,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _ExternalStage(ResourceSpec):
    name: ResourceName
    url: str
    owner: Role = "SYSADMIN"
    type: StageType = StageType.EXTERNAL
    storage_integration: str = None
    credentials: dict[str, str] = field(default_factory=None, metadata={"fetchable": False})
    encryption: dict[str, str] = field(default_factory=None, metadata={"fetchable": False})
    directory: dict[str, bool] = None
    tags: dict[str, str] = None
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.type != StageType.EXTERNAL:
            raise ValueError("Type must be EXTERNAL for _ExternalStage")
        valid_encryption_types = ["AWS_CSE", "AWS_SSE_S3", "AWS_SSE_KMS", "GCS_SSE_KMS", "AZURE_CSE", "NONE"]
        if self.encryption and self.encryption["type"] not in valid_encryption_types:
            raise ValueError(
                f"Invalid encryption type: {self.encryption.get('type')}. Must be one of {valid_encryption_types}."
            )


class ExternalStage(ResourceNameTrait, Resource):
    """
    -- External stage
    CREATE [ OR REPLACE ] [ { TEMP | TEMPORARY } ] STAGE [ IF NOT EXISTS ] <external_stage_name>
        externalStageParams
        directoryTableParams
        [ COMMENT = '<string_literal>' ]
        [ [ WITH ] TAG ( <tag_name> = '<tag_value>' [ , <tag_name> = '<tag_value>' , ... ] ) ]

    externalStageParams (for Amazon S3) ::=
        URL = { 's3://<bucket>[/<path>/]' | 's3gov://<bucket>[/<path>/]' }

        [ { STORAGE_INTEGRATION = <integration_name> } | { CREDENTIALS = ( {  { AWS_KEY_ID = '<string>' AWS_SECRET_KEY = '<string>' [ AWS_TOKEN = '<string>' ] } | AWS_ROLE = '<string>'  } ) } ]
        [ ENCRYPTION = ( [ TYPE = 'AWS_CSE' ] [ MASTER_KEY = '<string>' ] |
                        [ TYPE = 'AWS_SSE_S3' ] |
                        [ TYPE = 'AWS_SSE_KMS' [ KMS_KEY_ID = '<string>' ] ] |
                        [ TYPE = 'NONE' ] ) ]

    externalStageParams (for Google Cloud Storage) ::=
        URL = 'gcs://<bucket>[/<path>/]'
        [ STORAGE_INTEGRATION = <integration_name> ]
        [ ENCRYPTION = ( [ TYPE = 'GCS_SSE_KMS' ] [ KMS_KEY_ID = '<string>' ] | [ TYPE = 'NONE' ] ) ]

    externalStageParams (for Microsoft Azure) ::=
        URL = 'azure://<account>.blob.core.windows.net/<container>[/<path>/]'
        [ { STORAGE_INTEGRATION = <integration_name> } | { CREDENTIALS = ( [ AZURE_SAS_TOKEN = '<string>' ] ) } ]
        [ ENCRYPTION = ( [ TYPE = 'AZURE_CSE' ] [ MASTER_KEY = '<string>' ] | [ TYPE = 'NONE' ] ) ]

    externalStageParams (for Amazon S3-compatible Storage) ::=
        URL = 's3compat://{bucket}[/{path}/]'
        ENDPOINT = '<s3_api_compatible_endpoint>'
        [ { CREDENTIALS = ( AWS_KEY_ID = '<string>' AWS_SECRET_KEY = '<string>' ) } ]

    directoryTableParams (for Amazon S3) ::=
        [ DIRECTORY = ( ENABLE = { TRUE | FALSE }
                        [ REFRESH_ON_CREATE =  { TRUE | FALSE } ]
                        [ AUTO_REFRESH = { TRUE | FALSE } ] ) ]

    directoryTableParams (for Google Cloud Storage) ::=
        [ DIRECTORY = ( ENABLE = { TRUE | FALSE }
                        [ AUTO_REFRESH = { TRUE | FALSE } ]
                        [ REFRESH_ON_CREATE =  { TRUE | FALSE } ]
                        [ NOTIFICATION_INTEGRATION = '<notification_integration_name>' ] ) ]

    directoryTableParams (for Microsoft Azure) ::=
        [ DIRECTORY = ( ENABLE = { TRUE | FALSE }
                        [ REFRESH_ON_CREATE =  { TRUE | FALSE } ]
                        [ AUTO_REFRESH = { TRUE | FALSE } ]
                        [ NOTIFICATION_INTEGRATION = '<notification_integration_name>' ] ) ]
    """

    resource_type = ResourceType.STAGE
    props = Props(
        url=StringProp("url"),
        storage_integration=IdentifierProp("storage_integration"),
        encryption=PropSet(
            "encryption",
            Props(
                type=StringProp("type"),
                master_key=StringProp("master_key"),
                kms_key_id=StringProp("kms_key_id"),
            ),
        ),
        credentials=PropSet(
            "credentials",
            Props(
                aws_key_id=StringProp("aws_key_id"),
                aws_secret_key=StringProp("aws_secret_key"),
                aws_token=StringProp("aws_token"),
                aws_role=StringProp("aws_role"),
            ),
        ),
        directory=PropSet(
            "directory",
            Props(
                enable=BoolProp("ENABLE"),
                refresh_on_create=BoolProp("REFRESH_ON_CREATE"),
            ),
        ),
        tags=TagsProp(),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _ExternalStage

    def __init__(
        self,
        name: str,
        url: str,
        owner: str = "SYSADMIN",
        type: StageType = StageType.EXTERNAL,
        storage_integration: str = None,
        credentials: dict[str, str] = None,
        encryption: dict[str, EncryptionType] = None,
        directory: dict[str, bool] = None,
        tags: dict[str, str] = None,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        if directory is None:
            directory = {"enable": False}
        self._data: _ExternalStage = _ExternalStage(
            name=self._name,
            url=url,
            owner=owner,
            type=type,
            storage_integration=storage_integration,
            credentials=credentials,
            encryption=encryption,
            directory=directory,
            tags=tags,
            comment=comment,
        )


StageTypeMap = {
    StageType.INTERNAL: InternalStage,
    StageType.EXTERNAL: ExternalStage,
}


def _resolver(data: dict):
    return StageTypeMap[StageType(data["type"])]


Resource.__resolvers__[ResourceType.STAGE] = _resolver
